/* 
    AVR Spy v0.1
    AVR Spy Main File
    
    See avrspy_test.c for how to use it.
    
    Author:         rk
    Last Change:    Jul-24-2007
    Copyright:      Lg ins Land
*/

#include "avrspy.h"
#include <avr/pgmspace.h>
#include <avr/eeprom.h>

// !!!!!!!! change this for your C's SFR name table: !!!!!!!!!!!!
#include "res/table_iom8.h"
//#include "res/table_dummy.h" // empty dummy name table if FLASH memory short


/* Standard  Tool Functions (ADC, PWM, ...) */

/* ADC */
// VCC/5Volt as Ref; Left-Adjust (10bit to 16bit)
#define _ADC_REFS (0<<REFS1)|(1<<REFS0)|(1<<ADLAR)  
uint16_t ADC_initReadSingle(uint8_t channel) 
{
	ADMUX = channel|_ADC_REFS;
	uint8_t old=ADCSRA;
	
	// Den ADC aktivieren und Takt-Teilungsfaktor auf 64 stellen
	ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1);
	// 1. dummy readout
	SETBIT(ADCSRA,ADSC); // start conversion
	while(TESTBIT(ADCSRA,ADSC)); // warten bis fertig 
	// readout
	SETBIT(ADCSRA,ADSC); // start conversion
	while(TESTBIT(ADCSRA,ADSC)); // warten bis fertig 
	uint16_t v=ADC;
    
	ADCSRA=old;
    
	return v;
}

void timer1_init_pwm16bit(uint16_t init_value, uint16_t max_count, uint8_t prescale)
// prescale=1 : clk/1 ..
// prescale=4 : clk/256
// prescale=5 : clk/1024
{
    // COM1A1: clear at match
    TCCR1A = (1<<COM1A1)|(1<<WGM11);
    // CS10=1: Prescaler = 1   (16MHz/16bit: 244.1Hz; 16MHz/10bit=15.625kHz)
    // CS12=1: Prescaler = 256 (16MHz/16bit: 0.953674316Hz)
    TCCR1B = prescale|(1<<WGM12)|(1<<WGM13);
    // Interrupts fr Timer1 deaktivieren
    // Achtung: Auch die Interrupts fr die anderen Timer stehen in diesem Register
    TIMSK &= ~0x3c;
	
    ICR1H = max_count>>8;
    ICR1L = max_count;   // top: 16bit max. value
    
    OCR1AH = init_value>>8;  // 25% init 
    OCR1AL = init_value;
}
void timer1_set_pwm16bit(uint16_t v) 
{
    OCR1AH = v>>8;
    OCR1AL = v;
}
 

/* AVR Spy Core */ 

// replacements for sscanf parser
int parseAZaz09_(char** ps, char *x, int xmax)
{
	char *p=*ps,*p0;
	for(;*p==' ';p++); // eat white space
	p0=p;
	while ( (('A'<=*p)&&(*p<='Z'))
			|| (('a'<= *p)&&(*p <='z'))
			|| (('0'<= *p)&&(*p <='9'))
			|| *p=='_' ) {
		if (p-p0>=xmax) break;
		*x++ = *p++;
	}
	if (p==p0)
		return 0;
	*x=0;
	*ps=p;
	return 1;
}
int parseCmpI(char** ps, char *x)
// case insensitive string compare
{
	char *p=*ps,*p0;
	for(;*p==' ';p++); // eat white space
	p0=p;
	while (*p && tolower(*p)==tolower(*x)) {
		x++; p++;
	}
	if (*x || p==p0)
		return 0;
	*ps=p;
	return 1;
}
int parseHEX(char** ps, unsigned *x)
{
	char *p=*ps,*p0;
	unsigned v=0;
	for(;*p==' ';p++); // eat white space
	p0=p;
	for(;;p++) {
		switch (*p) {
		case '0':		case '1':		case '2':		case '3':
		case '4':		case '5':		case '6':		case '7':
		case '8':		case '9':
            v=v*16 + *p - '0';
			break;
		case 'a':		case 'b':		case 'c':
		case 'd':		case 'e':		case 'f':
            v=v*16 + *p - 'a' + 10;
			break;
		case 'A':		case 'B':		case 'C':
		case 'D':		case 'E':		case 'F':   
			v=v*16 + *p - 'A' + 10;
			break;
		default:
			goto _end;
		}
	}
_end:
	if (p==p0) return 0;
	*x=v;
	*ps=p;
	return 1;
}
int parseDEC(char** ps, unsigned *x)
{
	char *p=*ps,*p0;
	unsigned v=0;
	for(;*p==' ';p++); // eat white space
	p0=p;
	for(;(*p>='0' && *p<='9');p++) {
		v=v*10+ *p - '0';
	}
	if (p==p0) return 0;
	*x=v;
	*ps=p;
	return 1;
}

// non-interrupt UART init with 8bit,1stop
void spy_uart_init(unsigned UART_BUAD)
{
#if F_CPU < 2000000UL && defined(U2X)
  UCSRA = _BV(U2X);             /* improve baud rate error by using 2x clk */
  UBRRL = (F_CPU / (8UL * UART_BAUD)) - 1;
#else
  UBRRL = (F_CPU / (16UL * UART_BAUD)) - 1;
#endif
  UCSRB = _BV(TXEN) | _BV(RXEN); /* tx/rx enable */
}
inline int spy_putc(char c)
{
  if (c == '\n')
    spy_putc('\r');
  loop_until_bit_is_set(UCSRA, UDRE);
  UDR = c;
  return 0;
}
#define _putc spy_putc
int spy_puts(char *s)  // no auto line break \n
{
	for(;*s;s++) spy_putc(*s);
	return 0;
}
#define _puts spy_puts
// non-blocking get-char
inline int spy_getc(void)
// returns -1 if no char
{
	if (!TESTBIT(UCSRA, RXC)) return -1;
	//if (UCSRA & _BV(FE))
	//  return _FDEV_EOF;
	//if (UCSRA & _BV(DOR))
	//  return _FDEV_ERR;
	return UDR;
}
#define _getc spy_getc

// enable and call this in order have <stdio.h> functions puts,printf,
// gets,sscanf,... on the USART
#define SPY_STDIO
#ifdef SPY_STDIO
#include <stdio.h>
int spy_putchar(char c, FILE *stream) 
{
	return spy_putc(c);
}
int spy_getchar(FILE *stream)
{
	int c;
	while((c=spy_getc())==-1);
	return c;
}
FILE spy_uart_str = FDEV_SETUP_STREAM(spy_putchar, spy_getchar, _FDEV_SETUP_RW);
void spy_stdio(void)
{
    stdout = stdin = stderr = &spy_uart_str;
}
#endif


SPY_CB_FUNC *the_tool_callback=0;
void spy_init(SPY_CB_FUNC *tool_callback)
{
    the_tool_callback=tool_callback;
	spy_uart_init(9600);
#ifdef SPY_STDIO
	spy_stdio();
#endif
	spy_puts("AVR SPY:");
}

char spy_buf[20];
#define buf spy_buf
char spy_buf_history1[20]="";
uint8_t spy_buf_ix=0;

// non-blocking step function for driving the Spy
// returns 0 at 'q(uit)', <char> at open interaction, >=256 at command execution
int spy_step(void)
{
	char c,s[20], *p;
	int i;
	uint8_t b;
	uint16_t adr,w,bitpos;

	c=spy_getc();
	if (c==-1) return c;
	if (c=='\r') c='\n';
	// backspace delete
	if ((c==8||c=='\x7f') && spy_buf_ix) { 
		spy_buf_ix--; spy_putc(8);spy_putc(' ');spy_putc(8);
		return c;
	}
	// history scrolling
	if (c==0x41 && spy_buf_ix>=2 && spy_buf[spy_buf_ix-1]==0x5B && spy_buf[spy_buf_ix-2]==0x1B) {
		for(i=0;i<spy_buf_ix+2;i++) spy_putc(8);
		for(i=0;i<spy_buf_ix+2;i++) spy_putc(' ');
		for(i=0;i<spy_buf_ix+2;i++) spy_putc(8);
		strcpy(spy_buf,spy_buf_history1);
		spy_buf_ix=strlen(spy_buf);
		spy_puts(": ");spy_puts(spy_buf);
		return c;
	} else if (spy_buf_ix<sizeof(spy_buf)-1) {
		// normal add to buffer
		spy_buf[spy_buf_ix++]=c; 
		if (c>' ' && !(spy_buf_ix>=2 && spy_buf[spy_buf_ix-2]==0x1B)) // ESC
			spy_putc(c); // echo
	}
	if (c!='\n') return c;
	else  {
		puts("\r");
		spy_buf[spy_buf_ix++]=0;
		if (spy_buf_ix>1) {
			strncpy(spy_buf_history1,spy_buf,strlen(spy_buf)-1);
			spy_buf_history1[strlen(spy_buf)-1]=0;
		}
		spy_buf_ix=0; // reset buf for next edit cycle
		
		if (buf[0] == 'q')
			return 0;
		switch (buf[0])
		{
			
		case 'm': // RAM-Memory
			if ((p=buf+1) && *p++=='w' && parseHEX(&p,&adr) && *p++=='=' && parseHEX(&p,&w) )
			{
				*((uint16_t*)adr) = w;
				printf("RAM[%04X]=%04X\n", adr, w);
			} else if ((p=buf+1) && parseHEX(&p,&adr) && *p++=='=' && parseHEX(&p,&w) )
			{
				*((uint8_t*)adr) = b = w;
				printf("RAM[%04x]=%02X\n", adr, b);
			} else if ((p=buf+1) && *p++=='w' && parseHEX(&p,&adr))
			{
				printf("RAM[%04X]: %04X\n", adr, *((uint16_t*)adr));
			} else if ((p=buf+1) && parseHEX(&p,&adr))
			{
				printf("RAM[%04X]: %02X\n", adr, *((uint8_t*)adr));
			} else 
                printf("RAM %s unkown\n", buf+1);
			break;

		case 'p': // FLASH Memory
			if ((p=buf+1) && parseHEX(&p,&adr))
			{
				printf("FLASH[%04X]: %02X\n", adr, pgm_read_byte(adr));
			} else 
                printf("FLASH %s unkown\n", buf+1);
			break;
			
		case 'e': // EEPROM Memory
			if ((p=buf+1) && parseHEX(&p,&adr) && *p++=='=' && parseHEX(&p,&w) )
			{
				eeprom_write_byte((uint8_t*)adr, w);
				printf("EEP[%04x]: %02X\n", adr, eeprom_read_byte((uint8_t*)adr));
			} else if ((p=buf+1) && parseHEX(&p,&adr)) {
				printf("EEP[%04X]: %02X\n", adr, eeprom_read_byte((uint8_t*)adr));
			} else 
                printf("EEP %s unkown\n", buf+1);
			break;

		case 's': // SFR access
			if ( (p=buf+1) && parseAZaz09_(&p,s,sizeof(s)) && *p++=='=' && parseHEX(&p,&w) ) { //"%*c%20[a-zA-Z0-9]=%x"
				for (i=0; i<ARRAYLEN(aReg8Names); i++)
					if (strcasecmp_P(s,(char*)pgm_read_word(&aReg8Names[i]))==0) {
						adr=pgm_read_byte(&aReg8Adrs[i])+0x20;
						*((uint8_t*)adr) = w;
						printf("%s=%02X\n", s, *((uint8_t*)adr) );
						i=0;
						break;
					}
				if (i) for (i=0; i<ARRAYLEN(aReg16Names); i++)
					if (strcasecmp_P(s,(char*)pgm_read_word(&aReg16Names[i]))==0) {
						adr=pgm_read_byte(&aReg16Adrs[i])+0x20;
						*((uint16_t*)adr) = w;
						printf("%s=%04X\n", s, *((uint16_t*)adr) );
						i=0;
						break;
					}
				if (i) 
					printf("SFR %s unkown\n", s);
			} else 	if ((p=buf+1) && parseAZaz09_(&p,s,sizeof(s)) && *p++=='/' && parseHEX(&p,&bitpos) && *p++=='=' && parseHEX(&p,&w)) {
				for (i=0; i<ARRAYLEN(aReg8Names); i++)
					if (strcasecmp_P(s,(char*)pgm_read_word(&aReg8Names[i]))==0) {
						adr=pgm_read_byte(&aReg8Adrs[i])+0x20;
						WRITEBIT(*((uint8_t*)adr),bitpos,w);
						printf("%s=%02X\n", s, *((uint8_t*)adr) );
						i=0;
						break;
					}
				if (i) 
					printf("SFR %s unkown\n", s);
                
			} else if ((p=buf+1) && parseAZaz09_(&p,s,sizeof(s)))			{
				for (i=0; i<ARRAYLEN(aReg8Names); i++)
					if (strcasecmp_P(s,(char*)pgm_read_word(&aReg8Names[i]))==0) {
						adr=pgm_read_byte(&aReg8Adrs[i])+0x20;
						printf("%s: %02X\n", s, *((uint8_t*)adr) );
						i=0;
						break;
					}
				if (i) for (i=0; i<ARRAYLEN(aReg16Names); i++)
					if (strcasecmp_P(s,(char*)pgm_read_word(&aReg16Names[i]))==0) {
						adr=pgm_read_byte(&aReg16Adrs[i])+0x20;
						printf("%s: %04X\n", s, *((uint16_t*)adr) );
						i=0;
						break;
					}
				if (i) 
					printf("SFR %s unkown\n", s);
			} 
			break;
			
		case 'c': // blocking stop and 10x HEX terminal char echo
			for (i=0; i<10; i++) {
				loop_until_bit_is_set(UCSRA, RXC);
				b=UDR; 
				printf("x%02X ",b);
			}
			break;
			
		default:
			
			// Functions & Macros
			
			if ((p=buf) && parseCmpI(&p,"AD") && parseHEX(&p,&adr)) {
				printf("ADC[%d]: %u\n",adr, ADC_initReadSingle(adr));
				break;
			}
			if ((p=buf) && parseCmpI(&p,"PWM1=") && parseDEC(&p,&w)) {
				unsigned wmax=0xFFFF;
				unsigned prescale=4;  //
				if (*p++==',' && parseDEC(&p,&wmax));
				if (*p++==',' && parseDEC(&p,&prescale));
				prescale &= 7;
				timer1_init_pwm16bit(w,wmax,prescale);
				printf("PWM1 set.\n");
				break;
			}
            
            // Callback to User Tools
            
            if (!*the_tool_callback || !(*the_tool_callback)(buf)) 
				if  (buf[0]!='\n') {
					_puts("Unknown command: ");_puts(buf);
				}
			break; 

		} // of switch 1st char
		spy_puts("> ");
    } // of \n handler
	return 256+buf[0];
}

void spy_loop(void)
{
	spy_puts("\n> ");
	while (spy_step());

	printf("Bye.\n");
}
