/* 
   Shows how to use/attach the AVR Spy
*/

#include "avrspy.h"  
#include <stdio.h>

void delay_n(uint16_t t) // wait some time
{
  unsigned int i,j;
  for (j = 0; j < t; j++)
	for (i = 0; i < 1000; i++)
		asm volatile ("nop");
}

// a bunch of my extra tool commands:
int my_spy_tool_callback(char *cmdline)
{
	char *p;
	unsigned w;
	
    // Beep Tool: "BEEP=100"
	if ((p=cmdline) && parseCmpI(&p,"BEEP=") && parseDEC(&p,&w)) {
        puts("beep...");
		SETBIT(PORTD,PD7);  // Buzzer ON
		delay_n(w);         // wait
		CLRBIT(PORTD,PD7);  // Buzzer OFF
		return 1;           // command recognized
	}
    
    // Blink Tool: "Blink=5,200"     (1 or 2 parameters)
	if ((p=cmdline) && parseCmpI(&p,"BLINK=") && parseDEC(&p,&w)) {
        unsigned i,delay=100;
        if (*p++==',' && parseDEC(&p,&delay)); // optional 2nd parameter
        puts("blink...");
        for (i=0; i<w; i++) {
            INVBIT(PORTD,PD5); // Toggle LED1 
            delay_n(delay);       // wait
            INVBIT(PORTD,PD5); // Toggle LED1 
            delay_n(delay);       // wait
        }
		return 1; // command recognized
	}
    
	return 0; // command not recognized
}


int main (void)
{
    // Port Init
    SETBIT(DDRB, 1);  // PB1/OC1A-PWM as output
	DDRD  |= 0xE0;    // LED1, LED2, BEEPER on Pollin.de's ATMEL Eval-Board.
	PORTD = 0x20;
    
    // 'AVR SPY:' apears on UART (9600,8,n,1) - and we have stdio.h then
 	spy_init(my_spy_tool_callback); // 
 	//spy_init(0);  // that's how to not link a tool_callback

    puts("Spy-only loop until 'q(uit)' is pressed 1st time:");
    spy_loop();   
   
    puts("2nd life: my mainloop feeding spy with spy_step()");
    printf("> ");
    while (1) {
        unsigned int u;
        
        INVBIT(PORTD,PD6); // Toggle LED2 
        for(u=0;u<10000;u++) 
            spy_step();     // wait feeding the spy shell
        INVBIT(PORTD,PD6); // Toggle LED2
        for(u=0;u<10000;u++) 
            spy_step();     // wait feeding the spy shell
    }
}
